Artifact Overview
----

We briefly describe the correspondence between our Scala implementation and paper. The code is contained in the directory `oopsla19-code/src/main/scala/sai`.

`lang` contains the abstract syntax of the language (Section 2.1).
`monads` contains the monads implementation used throughout the paper.
`lattices` contains the implementation of lattice structures used in the paper (e.g., Section 5.1 and 6.1).
`lms` contains our extension to the Lightweight Modular Staging framework, e.g., data structures support, optimization in code generation and few bug fix (the original LMS library is not changed).
`evaluation` contains programs to reproduce the empirical evaluation (Section 8).
`interpreters` contains the main implementation of our four interpreters:
`Semantics.scala`, the generic interpreter (Section 2.3)
`Concrete.scala`, the concrete interpreter (Section 3)
`StagedConcrete.scala`, the staged concrete interpreter (Section 4)
`Abstract.scala`, the abstract interpreter (Section 5)
`StagedAbstract.scala`, the staged abstract interpreter (Section 6)

Reproduce the Empirical Evaluation
----

The reviewer may run

`cd ~/sai/oopsla19-code/; ./start_sbt`

to start the interactive build tool, which sets the JVM options.

To (re)compile the .scala files, run
	`clean`
and then
	`compile`
in sbt. 

To reproduce the table of result in Section 8, run

`runMain sai.evaluation.Evaluation`

in sbt. 

For each benchmark, we run 20 times for both unstaged and staged analyzer and take the median number to calculate the speedup.  The benchmark programs are contained under folder ~/sai/oopsla19-code/benchmarks. 

The console should output the result for each benchmark, including the AST size of benchmarks, the time used for each iteration, the mean and median number for both unstaged and staged version, and the calculated speedup. Note that the number obtained under a virtual machine may differ from what we reported in the paper, but should fall in the approximate range. The generated programs will also be written into external files, as indicated by the log. The whole process takes roughly 40 minutes to finish.

As a reference, the original data used in the paper is located at ~/sai/data/original_data.txt.

